<?php

namespace Drupal\dynamic_path_aliases;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Serialization\Exception\InvalidDataTypeException;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Serialization\Yaml;
use Drupal\dynamic_path_aliases\Form\SettingsForm;

/**
 * Provides dynamic path aliases manager.
 */
class DynamicPathsManager implements DynamicPathsManagerInterface {

  /**
   * The cache key prefix.
   */
  const CACHE_KEY = 'dynamic-path-aliases';

  /**
   * The cache backend.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected CacheBackendInterface $cache;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The config.
   *
   * @var array
   */
  protected array $config;

  /**
   * DynamicPathsManager constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache backend.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(ConfigFactoryInterface $configFactory, CacheBackendInterface $cache, EntityTypeManagerInterface $entityTypeManager) {
    $this->cache = $cache;
    $this->entityTypeManager = $entityTypeManager;
    $this->config = [];

    if ($config = $configFactory->get(SettingsForm::SETTINGS)->get('aliases')) {
      try {
        $this->config = array_filter(Yaml::decode($config));
      }
      catch (InvalidDataTypeException $e) {
        watchdog_exception('dynamic_path_aliases', $e);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function match(string $path, string $source, string $return): ?string {
    if ($cache = $this->getCache("$path:$source:$return")) {
      return $cache;
    }
    foreach ($this->config as $definition) {
      $pattern = preg_replace('/\\\\\{[\w_]+\\\}/', '([^\/]+)', preg_quote($definition[$source], '/'));
      if (preg_match("/^$pattern$/i", $path, $params)) {
        // Combine expected route slugs with matched parameters. All slugs
        // in the definition must have matching parameters.
        preg_match('/\{[\w_]+}/', $definition[$source], $slugs);
        array_walk($slugs, function (&$slug): void {
          $slug = str_replace(['{', '}'], NULL, $slug);
        });
        array_shift($params);
        $params = array_combine($slugs, $params);
        if (count(array_filter($params)) !== count($slugs)) {
          continue;
        }

        // Verify route entities and finalize return path.
        $skip = FALSE;
        $returnPath = $definition[$return];
        $tags = [];
        foreach ($params as $key => $value) {
          if (isset($definition['entities'][$key])) {
            try {
              if ($entity = $this->entityTypeManager->getStorage($definition['entities'][$key]['type'])->load($value)) {
                if (!isset($definition['entities'][$key]['bundle']) || in_array($entity->bundle(), $definition['entities'][$key]['bundle'], TRUE)) {
                  $returnPath = str_replace("{{$key}}", $entity->id(), $returnPath);
                  $tags = array_merge($tags, $entity->getCacheTags());
                  continue;
                }
              }
            }
            catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
              watchdog_exception('dynamic_path_aliases', $e);
            }
          }
          // Aborts when required entities are missing.
          $skip = TRUE;
        }

        if ($skip) {
          continue;
        }
        $this->setCache("$path:$source:$return", $returnPath, $tags);
        return $returnPath;
      }
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getPathByAlias(string $path): ?string {
    return $this->match($path, 'alias', 'path');
  }

  /**
   * {@inheritdoc}
   */
  public function getAliasByPath(string $path): ?string {
    return $this->match($path, 'path', 'alias');
  }

  /**
   * {@inheritdoc}
   */
  public function getCache(string $key): ?string {
    if ($cache = $this->cache->get(static::CACHE_KEY . ":$key")) {
      return $cache->data;
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function setCache(string $key, string $value, array $tags = []): void {
    $this->cache->set(static::CACHE_KEY . ":$key", $value, Cache::PERMANENT, $tags);
  }

}
